#include "Graph.h"

using namespace verifier;
using namespace collection;
using namespace shared;
using namespace std;
using namespace help;

/*
 *	Creates a graph with 1+orbit1+orbit2 vertices.
 */
Graph::Graph( unsigned int orbit1, unsigned int orbit2 ) : orbit1(orbit1), orbit2(orbit2), vertices(orbit1+orbit2+1) {
	flag = false;

	A = ArrayPool::fetch( vertices ); // A = new int[ vertices ];
	memset( A, 0, vertices*sizeof(int) );

	anonymousEdges = ArrayPool::fetch( vertices ); // anonymousEdges = new int[ vertices ];
	memset( anonymousEdges, 0, vertices*sizeof(int) );
}

/*
 *	Creates a copy of another graph.
 */
Graph::Graph( const Graph& other ) {
	assert( other.anonymousEdges != 0 );
	vertices = other.vertices;

	if ( vertices == 0 ) {
		// Copied an invalid graph
		return;	
	}

	// Copy adjacency matrix
	A = ArrayPool::fetch( vertices ); // A = new int[ vertices ];
	memcpy( A, other.A, vertices*sizeof(int) );

	// Copy anonymous edges
	anonymousEdges = ArrayPool::fetch( vertices ); // anonymousEdges = new int[ vertices ];
	memcpy( anonymousEdges, other.anonymousEdges, vertices*sizeof(int) );

	orbit1 = other.orbit1;
	orbit2 = other.orbit2;
	flag = other.flag;
}

/*
 *	Read a graph from a file (piped through a ByteReader object)
 */
Graph::Graph(ByteReader& reader) {
	int numOrbits = reader.getNumber(1);

	if ( numOrbits == 0 ) {
		// Invalid graph read. This might happen due to padding appended
		// to the end of a binary file.
		vertices = 0;	
		A = NULL;
		return;
	}

	int numOrbit1Vertex = reader.getNumber(3) + 1;
	this->orbit1 = numOrbit1Vertex;

	int maxNumVertex = numOrbit1Vertex + 12;

	bool connectionsOrbit1[numOrbit1Vertex][maxNumVertex];
	for( int i = 0; i < numOrbit1Vertex; ++i ) {
		for( int j = 0; j < maxNumVertex; ++j ) {
			connectionsOrbit1[i][j] = false;
		}
	}

	//Orbit 1
	int numOrbit2Vertex = 0;
	for( int vertex = 0; vertex < numOrbit1Vertex; ++vertex ){
		int numConnectionsInOrbit1 = reader.getNumber(3);

		for ( int i = 0; i < numConnectionsInOrbit1; ++i ){
			int connectedWith = reader.getNumber(3);
			connectionsOrbit1[vertex][connectedWith] = true;
		}

		int numConnectionsInOrbit2 = reader.getNumber(3);

		for ( int i = 0; i < numConnectionsInOrbit2; ++i ){
			int connectedWith = reader.getNumber(4);
			connectionsOrbit1[vertex][connectedWith + numOrbit1Vertex] = true;

			// Take the maximum.
			if( (connectedWith + 1) > numOrbit2Vertex)
				numOrbit2Vertex = connectedWith + 1;
		}

	}

	//Orbit 2
	this->orbit2 = numOrbit2Vertex;

	// initialize arrays
	bool connectionsOrbit2[numOrbit2Vertex][numOrbit2Vertex];
	int anonymousEdgesOrbit2[numOrbit2Vertex];
	for( int i = 0; i < numOrbit2Vertex; ++i ){
		for( int j = 0; j < numOrbit2Vertex; ++j ){
			connectionsOrbit2[i][j] = false;
		}
		anonymousEdgesOrbit2[i] = 0;
	}

	// read orbit2 edges
	for( int vertex = 0; vertex < numOrbit2Vertex; ++vertex ){
		int numConnectionsInOrbit2 = reader.getNumber(3);

		for ( int i = 0; i < numConnectionsInOrbit2; ++i ){
			int connectedWith = reader.getNumber(4);
			connectionsOrbit2[vertex][connectedWith] = true;
		}

		int anonymous = reader.getNumber(3);
		anonymousEdgesOrbit2[vertex] = anonymous;
	}

	// Set all the information in the matrix.
	int totalNumVertex = numOrbit1Vertex + numOrbit2Vertex + 1;
	vertices = totalNumVertex;

	A = ArrayPool::fetch( totalNumVertex ); // A = new int[ totalNumVertex ];
	for ( int i = 0; i < totalNumVertex; ++i )
		A[i] = 0;

	// Set the connections from orbit1
	for(int i = 1; i <= numOrbit1Vertex; ++i){
		setEdge( 0, i );
	}

	for ( int i = 0; i < numOrbit1Vertex; ++i){
		for ( int j = 0; j < (numOrbit1Vertex + 12); ++j){
			if ( connectionsOrbit1[i][j] ){
				setEdge( i + 1, j + 1 );
			}
		}
	}

	// Set the connections for orbit2
	for ( int i = 0; i < numOrbit2Vertex; ++i){
		for ( int j = 0; j < numOrbit2Vertex; ++j){
			if ( connectionsOrbit2[i][j] ){
				setEdge( ( i + numOrbit1Vertex + 1 ), ( j + numOrbit1Vertex + 1 ) );
			}
		}
	}

	// Set the anonymous edges
	anonymousEdges = ArrayPool::fetch( totalNumVertex ); // anonymousEdges = new int[ totalNumVertex ];

	for ( int i = 0; i <= numOrbit1Vertex; ++i){
		anonymousEdges[i] = 0;
	}

	for ( int i = 0; i < numOrbit2Vertex; ++i){
		anonymousEdges[i + numOrbit1Vertex + 1] = anonymousEdgesOrbit2[i];
	}

	reader.flush();
	flag = false;
}

/*
 *	Destructor
 */
Graph::~Graph() {
	if ( vertices > 0 ) {
		if ( A ) {
			ArrayPool::recycle( A, vertices ); // delete[] A;
			A = 0;
		}

		if ( anonymousEdges ) {
			ArrayPool::recycle( anonymousEdges, vertices ); // delete[] anonymousEdges;
			anonymousEdges = 0;
		}
	}
}

/*
 *  Returns the number of vertices of this graph.
 */
int Graph::size() const {
	return vertices;
}

/*
 *  Connects vertices (v,w). Does not check whether edge already exists.
 */
void Graph::setEdge( unsigned int v, unsigned int w ) {
	assert( v < vertices );
	assert( w < vertices );

	A[v] = A[v] | (1 << w);
	A[w] = A[w] | (1 << v);

}

/*
 *	Returns whether v is adjacent to w.
 */
bool Graph::adjacent( unsigned int v, unsigned int w ) const {
	assert( v < vertices && w < vertices );
	//assert( A->isSet( v, w ) == A->isSet( w, v ) ); // assert symmetry

	return ( (A[v] >> w) & 1 );
}

/*
 *	Returns the degree if vertex v, including anonymous edges.
 */
unsigned int Graph::getDegree( unsigned int v ) const {
	assert( v < vertices );

	int row = A[v];
	unsigned int res = 0;

	// Count the number of set bits in 'row'
	for( ; row; ++res ) {
		row &= row - 1;
	}

	return res + anonymousEdges[v];
}

/*
 *	Returns the number of anonymous edges v has.
 */
int Graph::getAnonymousEdges( unsigned int v ) const {
	assert( v < vertices );
	return anonymousEdges[ v ];
}

/*
 *	Sets the amount of anonymous edges for a vertex v.
 */
void Graph::setAnonymousEdges( unsigned int v, unsigned int edges ) {
	assert( v < vertices );

	anonymousEdges[ v ] = edges;
}

/*
 *	Copies graph into this one.
 */
Graph& Graph::operator= ( const Graph& other ) {
	assert( this != &other );
	
	if ( orbit1 != other.orbit1 || orbit2 != other.orbit2 ) {
		if ( A )
			delete[] A;
		if ( anonymousEdges )
			delete[] anonymousEdges;

		A = ArrayPool::fetch( other.vertices ); // A = new int[ vertices ];
		anonymousEdges = ArrayPool::fetch( other.vertices ); // anonymousEdges = new int[ vertices ];
	}

	vertices = other.vertices;
	orbit1 = other.orbit1;
	orbit2 = other.orbit2;
	flag = other.flag;

	// Copy adjacency matrix
	memcpy( A, other.A, vertices*sizeof(int) );

	// Copy anonymous edges
	memcpy( anonymousEdges, other.anonymousEdges, vertices*sizeof(int) );

	return *this;
}

/*
 *	Test for (simple) equality of two graphs.
 */
bool Graph::operator== ( const Graph& other ) const {
	if( orbit1 != other.orbit1 || orbit2 != other.orbit2 ) return false;
	assert( vertices == other.vertices );

	for ( int i = 0; i < vertices; ++i ) {
		// check neighbourhood and anonymous edges of current vertice
		if ( A[i] != other.A[i] || anonymousEdges[i] != other.anonymousEdges[i] )	
			return false;
	}

	return true;
}

/*
 *	Swaps the rows and columns of vertices a and b in the adjacency matrix.
 */
void Graph::swapVertices( unsigned int a, unsigned int b ) {

	unsigned int temp = 0;

	// Swap rows
	temp = A[a];
	A[a] = A[b];
	A[b] = temp;


	for (int i = 0; i < vertices; i++){
		temp = ((A[i] >> a) ^ (A[i] >> b)) & 1;
		A[i] ^= ((temp << a) | (temp << b));
	}

	/* // was:
	// Swap columns
	int mask = ~(0 | (1<<a) | (1<<b));	
	for (int i = 0; i < vertices; i++){
	int posa = adjacent(i, a) ? 1 : 0;
	int posb = adjacent(i, b) ? 1 : 0;
	A[i] = (A[i] & mask) | (posa<<b) | (posb<<a);
	}
	*/

	// Swap anonymous edges
	temp = anonymousEdges[a];
	anonymousEdges[a] = anonymousEdges[b];
	anonymousEdges[b] = temp;
}

void Graph::deleteVertex(int v){
	A[v] = 0;

	for (int i = 0; i < vertices; i++)
		A[i] = A[i] & ~(1 << v);

	anonymousEdges[v] = 0;

}

/*
 *	Writes this graph into an output file stream.
 */
void Graph::encodeToFile( ofstream* file ){
	ByteBuffer encoded;

	// Write #orbits (always 2)
	encoded.putBits(1, 1);

	// Write the number of orbit1 vertices.
	encoded.putBits( orbit1-1, 3 );

	// Write vertices of orbit1.
	List<char> connected;
	for ( int vertex = 1; vertex <= orbit1; ++vertex ){

		for ( int w = 1; w <= orbit1; ++w ){
			if( this->adjacent(vertex, w) )
				connected.push_back(w);
		}

		// How many vertices in orbit1 are connected with 'vertex'?
		encoded.putBits( connected.size(), 3);

		// Which ones?
		for ( List<char>::iterator i = connected.begin(); i != connected.end(); i++ ) {
			encoded.putBits( (*i)-1, 3 );
		}

		connected.clear();

		// How many vertices in orbit2 are connected to 'vertex'?
		for ( int w = orbit1 + 1; w < vertices; ++w ){
			if( this->adjacent(vertex, w) )
				connected.push_back(w);
		}

		// How many vertices in orbit2 are connected with 'vertex'?
		encoded.putBits( connected.size(), 3 );

		// Which ones?
		for ( List<char>::iterator i = connected.begin(); i != connected.end(); i++ ) {
			encoded.putBits( (*i)-1-orbit1, 4 );
		}

		connected.clear();
	}


	// Write vertices of orbit2.
	for ( int vertex = orbit1 + 1; vertex < vertices; ++vertex ){

		for ( int w = orbit1 + 1; w < vertices; ++w ){
			if( this->adjacent(vertex, w) )
				connected.push_back(w);
		}

		// How many vertices in orbit1 are connected with 'vertex'?
		encoded.putBits( connected.size(), 3 );

		// Which ones?
		for ( List<char>::iterator i = connected.begin(); i != connected.end(); i++ ) {
			encoded.putBits( (*i)-orbit1-1, 4 );
		}

		connected.clear();

		// How many anonymous edges does 'vertex' have?
		encoded.putBits( this->getAnonymousEdges(vertex), 3 );

	}

	// Write ByteBuffer to file.
	encoded.writeTo(*file);
}

/*
 *	Prints this graph to the standart output stream (for debugging).
 */
void Graph::print(){
	stringstream res( stringstream::in | stringstream::out );

	res << "   ";
	for( int i = 0; i < vertices; ++i ) {
		res << i << " "; 
	}
	res << "\n\n";

	for( int y = 0; y < vertices; ++y ) { 
		res << y << "  ";
		for( int x = 0; x < vertices; ++x ) { 
			res << ((A[y] >>  x) & 1   ? "1" : ".") << " ";
		}
		res << "\n";
	}

	res << "\n";
	for( int i = 0; i < vertices; ++i ) {
		res << getDegree( i ) << " ";
	}

	res << "\n";
	for( int i = 0; i < vertices; ++i ) {
		res << getAnonymousEdges( i ) << " ";
	}

	cout << res.str() << "\n\n";
}

branching Graph::branchWithMirrors() {
	cout << "Branching with mirrors" << endl;
	List<int>* mirrors = this->mirrors();

	// Create copies to branch on
	Graph takeRoot(*this);
	Graph leaveRoot(*this);

	// Branch: take root-node, delete its neighbours
	PRINT( "Including root into MIS" );
	for (int i = 0; i <= orbit1; i++)
		takeRoot.deleteVertex(i); // Deletes roots-node and first orbit

	bool changed;
	do {
		changed  = false;
		changed = takeRoot.domination() || changed;
		changed = takeRoot.folding() || changed;
		changed = takeRoot.cleaning() || changed;
	} while( changed );

	takeRoot.print();


	// Branch: exclude root-node and its mirrors from MIS
	PRINT( "Excluding root and its mirrors from MIS" );
	leaveRoot.deleteVertex(0);
	cout << "Mirrors: ";
	for (List<int>::iterator i = mirrors->begin(); i != mirrors->end(); i++) {
		leaveRoot.deleteVertex( *i );
		cout << (*i) << " ";
	}
	cout << endl;

	// Apply reduction rules
	do {
		changed = false;
		changed = leaveRoot.domination() || changed;
		changed = leaveRoot.folding() || changed;
		changed = leaveRoot.cleaning() || changed;
	} while( changed );

	leaveRoot.print();

	// Return resulting branching vectors
	return calculateBranching( takeRoot, leaveRoot );
}

branching Graph::branchWithSatellites(){
	cout << "Branching with satellites" << endl;
	List<int>* satellites = this->satellites();

	// Create copies to branch on
	Graph takeRoot(*this);
	Graph leaveRoot(*this);

	// Branch: take root-node and its satellites into MIS, delete resp. neighbours
	PRINT( "Including root and satellites into MIS" );
	for (int i = 0; i <= orbit1; i++)
		takeRoot.deleteVertex(i); // Deletes roots-node and first orbit

	cout << "Satellites: ";
	List<int> remove;	// Create list with nodes that should be deleted
	for (List<int>::iterator i = satellites->begin(); i != satellites->end(); i++){
		for (int j = orbit1 + 1; j < vertices; j++) {	// First orbit can be ignored, it is already deleted 
			if ( this->adjacent((*i), j) )
				remove.push( j );
		}
		remove.push( *i );
		cout << (*i) << " ";
	}
	cout << endl;

	cout << "Satellite neighbours: ";
	for (List<int>::iterator i = remove.begin(); i != remove.end(); i++) {
		takeRoot.deleteVertex( *i );
		cout << (*i) << " ";
	}
	cout << endl;
	
	bool changed;
	do {
		changed  = false;
		changed = takeRoot.domination() || changed;
		changed = takeRoot.folding() || changed;
		changed = takeRoot.cleaning() || changed;
	} while( changed );
		
	takeRoot.print();


	// Branch: exclude root-node from MIS, delete it
	PRINT( "Excluding root from MIS" );
	leaveRoot.deleteVertex(0);

	// Apply reduction rules
	do {
		changed  = false;
		changed = leaveRoot.domination() || changed;
		changed = leaveRoot.folding() || changed;
		changed = leaveRoot.cleaning() || changed;
	} while( changed );

	leaveRoot.print();

	// Return resulting branching vectors
	return calculateBranching( takeRoot, leaveRoot );
}

/*!
 *	Calculates and returns the mirrors of the root node
 */
List<int>* Graph::mirrors() {
	List<int>* mirrors = new List<int>();
	for(int mirror = orbit1 + 1; mirror < vertices; mirror++ ){
	// !!!!!!!!!!TODO!!!!!!!!! 
	//	if (anonymousEdges[mirror] != 0)
	//		continue;

		bool clique = true;
		for (int i = 1; i <= orbit1 && clique; i++){
			if (this->adjacent(i, mirror))
				continue;

			for (int j = i+1; j <= orbit1 && clique; j++){
				if (this->adjacent(j, mirror))
					continue;

				clique = clique && this->adjacent(i, j);
			}
		}

		if (clique)
			mirrors->push(mirror);
	}
	return mirrors;
}

/*!
 *	Calculates and returns the mirrors of the root node
 */
List<int>* Graph::satellites() {
	// If we find a vertex in orbit1 that is only connected to one orbit2
	// vertex, then this vertex in orbit2 is a satellite.
	List<int>* satellites = new List<int>();
	bool satellite[orbit2];
	for (int i = 0; i < orbit2; i++)
		satellite[i] = false;

	for (int i = 1; i <= orbit1; i++){

		int counter = 0;
		int last = 0;
		for (int j = orbit1 + 1; j < vertices; j++){
			if (this->adjacent(i, j)){
				counter++;
				last = j;
			}
		}

		if (counter == 1)
			satellite[last - (orbit1 + 1)] = true;

	}

	for (int i = 0; i < orbit2; i++){
		if (satellite[i])
			satellites->push(i+orbit1+1);
	}

	return satellites;
}


/*!
 *	Folds all (allowed) degree-2 vertices
 */
bool Graph::folding() {
	bool didSomething = false;
	for (int v = 0; v < vertices; v++){
		if (getAnonymousEdges(v) > 0)
			continue;

		if (getDegree(v) == 2) {
			int a = -1;
			while( !adjacent(++a, v) );
			int b = a;
			while( !adjacent(++b, v) );

			// Assert that both vertices are NOT adjacent
			if  ( adjacent( a, b ) )
				continue;

			// Assert that not both nodes have anonymous edges
			if ( anonymousEdges[a] != 0 && anonymousEdges[b] != 0 )
				continue;

			// Convert one of the vertices to the folden one.
			for (int i = 0; i < vertices; i++){
				if (adjacent(i, a))
					setEdge(b, i);
			}

			anonymousEdges[b] += anonymousEdges[a];

			deleteVertex(v);
			deleteVertex(a);
			cout << "Folded " << a << " and " << b << " over " << v << endl;
			didSomething = true;
		}
	}
	return didSomething;
}

/*!
 *	Removes all dominating vertices
 */
bool Graph::domination() {
	bool didSomething = false;
	for ( int i = 0; i < vertices; i++){
		for ( int j = 0; j < vertices; j++){

			if (!adjacent(i,j) || anonymousEdges[i] > 0 || anonymousEdges[j] > 0 )
				continue;

			if (A[i] & A[j] == A[j]) {
				cout << i << " dominates " << j << ", removing it";
				deleteVertex(i);
				didSomething = true;
			}
		}
	}
	return didSomething;
}

/*!
 *	Removes all degree-one vertices
 */
bool Graph::cleaning() {
	bool didSomething = false;
	for ( int i = 0; i < vertices; i++){
		if ( getDegree( i ) == 1 ) {
			cout << "Deleting " << i << endl;
			deleteVertex(i);
			i = 0;	// once again
			didSomething = true;
		}
	}
	return didSomething;
}

/*!
 *	Calculate the branching vector.
 */

branching Graph::calculateBranching( Graph& takeIt, Graph& leaveIt ) {
	branching res;

	// Populate vectors
	for ( int i = 0; i < vertices; ++i ) {
		assert( getDegree(i) != 0 );
		res.takeIt[ getDegree(i) ]++;
		res.leaveIt[ getDegree(i) ]++;

		res.takeIt[0] += getAnonymousEdges(i);
		res.leaveIt[0] += getAnonymousEdges(i);
	}

	// Calculate differences
	for ( int i = 0; i < vertices; ++i ) {
		if( takeIt.getDegree(i) == 0 ) continue;	// Vertex is deleted
		res.takeIt[ takeIt.getDegree(i) ]--;
		res.takeIt[ 0 ] -= takeIt.getAnonymousEdges(i);
	}

	for ( int i = 0; i < vertices; ++i ) {
		if( leaveIt.getDegree(i) == 0 ) continue;	// Vertex is deleted
		res.leaveIt[ leaveIt.getDegree(i) ]--;
		res.leaveIt[ 0 ] -= leaveIt.getAnonymousEdges(i);
	}	

	return res;
}

/* Unifies the graph representation by sorting the vertices inside their respective orbits
 * by degree (ascending).
 */
void Graph::unify() {
	// Store degree in separate array to avoid calling getDegree() every time.
	int* degree = ArrayPool::fetch( 1+orbit1+orbit2 );
	for( int i = 0; i < 1+orbit1+orbit2; ++i ) {
		degree[ i ] = getDegree( i );
	}

	// Sort orbit1 (selection sort)
	for( int i = 1; i <= orbit1; ++i ) {
		int min = i;

		// Find minimal element inside (i, orbit1]
		for( int k = i+1; k <= orbit1; ++k ) 
			if( degree[ k ] < degree[ min ] ) min = k;
		if ( i == min ) continue;

		// Swap degrees
		int temp = degree[i];
		degree[i] = degree[min];
		degree[min] = temp;

		// Swap vertices
		swapVertices( i, min );
	}	


	// Sort orbit2 (selection sort)
	for( int i = orbit1+1; i <= orbit1+orbit2; ++i ) {
		int min = i;

		// Find minimal element inside (i, orbit1+orbit2]
		for( int k = i+1; k <= orbit1+orbit2; ++k ) 
			if( degree[ k ] < degree[ min ] ) min = k;
		if ( i == min ) continue;

		// Swap degrees
		int temp = degree[i];
		degree[i] = degree[min];
		degree[min] = temp;

		// Swap vertices
		swapVertices( i, min );
	}

	ArrayPool::recycle( degree, 1+orbit1+orbit2 );
}
