#include "GraphGenerator.h"
#include "HelpFunctions.h"

using namespace help;
using namespace collection;
using namespace verifier;
using namespace std;

/*!
 * Sets the parameters to the given values. If this constructor is used, full isomorphic test
 * is enabled.
 *
 * @param orbit1 The Number of vertex the generated Graphs should have in the first orbit.
 * @param orbit2 The Number of vertex the generated Graphs should have in the second orbit.
 * @param minDegree The minimal degree that a vertex should have.
 * @param maxDegree The maximal degree that a vertex may have. 
 * @sa GraphGenerator::GraphGenerator( unsigned int orbit1, unsigned int orbit2, unsigned int minDegree, unsigned int maxDegree, unsigned int maxIso )
 */
GraphGenerator::GraphGenerator( unsigned int orbit1, unsigned int orbit2, unsigned int minDegree, unsigned int maxDegree, unsigned int parallelChoice, unsigned int maxNumberEntries )
    : orbit1(orbit1), orbit2(orbit2), minDegree(minDegree), maxDegree(maxDegree), 
      knownGraphsBase( orbit1, orbit1, orbit2, orbit2, maxNumberEntries ),
      knownGraphsComplete( orbit1, orbit1, orbit2, orbit2, maxNumberEntries ),
      knownGraphsAddEdges( orbit1, orbit1, orbit2, orbit2, maxNumberEntries ), 
      parallelChoice( parallelChoice ), bufferFlag(false) {

    totalVertices = 1 + orbit1 + orbit2;

    ostringstream temp;
    temp << "temp_graphs_" << orbit1 << "_" << orbit2 << "_" << minDegree << "_" << maxDegree;

    if (parallelChoice > 0 )
	temp << "_" << parallelChoice;

    filename = temp.str();
}

/*!
 * This is central function of this class. It generates all possible graphs with the number of
 * vertex in orbit 1 and orbit 2 and degrees between minDegree and maxDegree given in the constructor.
 * This task is done by the method complete(Graph* graph, int minDegree).
 * In a last step, every so obtained graph is filled up with further edges (if possible) so that every possible valid
 * combination of edges is finally generated.
 * 
 * @param results A pointer to a List of Graphs in which the resulting graphs will be stored
 */
void GraphGenerator::generateAllGraphs( ofstream* graphfile ){
    graphCounter = 0;
    PRINT("Generating graphs with orbits (" << orbit1 << ", " << orbit2 << ") and degree [" << minDegree << ", " << maxDegree << "]");

    ofstream* out;

    // A. Base graphs
    // Generate base graphs (full isomorphy check)
    PRINT( "Phase A: Creating base graphs.");
    PRINT_TIME;
    PRINT( "" );

    List<Graph*>* baseGraphs = new List<Graph*>();
    generateBaseGraphs( baseGraphs );
    // Clear the Trie used for isomorphism test while generating the base graphs.
    knownGraphsBase.clear();

    // Check whether there are base graphs at all.
    if ( baseGraphs->size() == 0 ) {
	delete baseGraphs;
	PRINT( "No base graphs found. Done." );
	return;
    }

    // B. Complete base graphs
    // Complete base graphs with edges, so that the degree of every vertice 
    // is between minDegree and maxDegree
    PRINT( "Phase B: Completing up to minimal degree.");
    PRINT_TIME;
    PRINT( "" );

    int numberBaseGraphs = baseGraphs->size();

    int j = deleteFinishedBaseGraphs(baseGraphs);
    while( baseGraphs->size() > 0 ) {
	Graph* graph = baseGraphs->peek();
	baseGraphs->pop();

	string completeFileName = getCompleteFileName(j);
	out = new ofstream( completeFileName.c_str(), ios::binary );

	if ( parallelChoice == 0 || parallelChoice == j ) {
	    PRINT("Completing BaseGraph number " << j);
	    PRINT_TIME;

	    List<Graph*>* onlyOrbit1 = new List<Graph*>();
	    completeNodeFirstOrbit(0, graph, onlyOrbit1);

	    for (List<Graph*>::iterator i = onlyOrbit1->begin(); i != onlyOrbit1->end(); i++ ) {
		complete( *i, out );
		delete *i;
	    }

	    delete onlyOrbit1;
	    
	}

	//delete graph;
	j++;

	// Write last graphs from buffer
	flushBuffer( out );

	out->close();
	delete out;
    }

    // Write dummy-file to indicate that all base-graphs have been generated successfully
    writeDummyFile();

    delete baseGraphs;

    knownGraphsComplete.clear();	// free up some space

    // C. Add more edges
    // Continue adding edges until every possible combination is created.
    PRINT( "Phase C: Adding edges.");
    PRINT_TIME;
    PRINT( "" );

    for (int i = 0; i < numberBaseGraphs; i++){
	int counter = 0;

	// Open next file
	string currentTempFile = getCompleteFileName(i);
	ByteReader in((char*)currentTempFile.c_str());

	while(!in.end()){
	    Graph* current = new Graph(in); // This graph is deleted inside addPossibleEdges(..)

	    // Check if invalid graph was read (might happen due to padding)
	    if ( current->vertices == 0 ) {
		delete current;
		break;
	    }

	    addPossibleEdges( current, 1, graphfile );

	    if ( (counter % 10000) == 0){
		DEBUG(counter);
		PRINT_TIME;
	    }
	    counter++;
	}

	// D. Finish & clean up
	// write graphs which are still in the buffer
	flushBuffer( graphfile );
    }

    knownGraphsAddEdges.clear();	// free up some space

    PRINT("Done, generated " << graphCounter << " graphs." );	
    PRINT_TIME;
}

void GraphGenerator::generateBaseGraphs( List<Graph*>* results ){
    Graph* seed = new Graph( orbit1, orbit2 ); 

    // create root node <-> orbit1 edges
    for ( int i = 1; i <= orbit1; ++i )
	seed->setEdge( 0, i );

    generateBaseGraphs( seed, results );
}

void GraphGenerator::generateBaseGraphs( Graph* graph, List<Graph*>* results ){
    if ( !knownGraphsBase.contains( *graph ) ) {
	knownGraphsBase.insert( *graph );
    } else {
	delete graph;
	return;
    }

    //Search for an invalid orbit1 vertex
    bool connected = true;
    unsigned int v = 0;
    while( v < orbit1 && connected ) {
	v++;
	connected = false;
	for ( unsigned int w = orbit1 + 1; w <= orbit1 + orbit2; ++w ) {
	    if ( graph->adjacent( v, w ) ) {
		connected = true;
		break;
	    }
	}
    }

    if ( !connected ) { 
	// A dangling vertex was found in orbit1.
	// Consider all possibilities to connect it with vertices from orbit2 
	// Note: looping backwards has shown to be a tiny bit faster
	for ( unsigned int w = orbit1 + orbit2; w >= orbit1 + 1; --w ) {
	    if ( !graph->adjacent( v, w ) && graph->getDegree(w) < maxDegree && graph->getDegree(v) < maxDegree ) {
		Graph* evolved = new Graph( *graph );
		evolved->setEdge( v, w );
		generateBaseGraphs( evolved, results );
	    }
	}
    } else { 
	// No invalid vertex was found in orbit1
	// Search for an invalid vertex in orbit2
	// Note: looping backwards has shown to be a tiny bit faster
	v = orbit1;
	connected = true;
	while( v < orbit1 + orbit2  && connected ) {
	    v++;
	    connected = false;
	    for ( unsigned int w = 1; w <= orbit1; ++w ) {
		if ( graph->adjacent( v, w ) ) {
		    connected = true;
		    break;
		}
	    }
	}

	if ( connected == false ) { 
	    // An unconnected vertex was found in orbit2
	    for ( unsigned int w = orbit1; w >= 1; --w ) {
		if ( !graph->adjacent( v, w ) && graph->getDegree(w) < maxDegree && graph->getDegree(v) < maxDegree ) {
		    Graph* evolved = new Graph( *graph );
		    evolved->setEdge( v, w );
		    generateBaseGraphs( evolved, results );
		}
	    }

	} else { 
	    // The graph is a valid base graph
	    results->push_back( new Graph(*graph) );
	}

    }

    delete graph;
}

void GraphGenerator::completeNodeFirstOrbit( int v, Graph* graph, List<Graph*>* out ){

    // Check whether graph is complete
    if ( v > orbit1 ) {	
	// THOU SHALL NOT TEST WHETHER THIS GRAPH IS INSIDE KNOWNGRAPHSADDEDGES, FOR
	// IT WILL SMITE THE CORRECTNESS OF YOUR PROGRAM!
	out->push( new Graph( *graph ) );
	delete graph;

	return;
    }

    // Check whether further edges can be added, that is, the 
    // maximal Degree is not yet reached.
    if ( graph->getDegree(v) < maxDegree ) {
	// Try all possible edges (v,w) 
	for ( int w = 1; w <= orbit1; ++w ) {
 
	    if ( v != w && (graph->getDegree(w) < maxDegree) ) {

		// Edge is possible, copy graph and add it. 
		Graph* evolved = new Graph( *graph );
		evolved->setEdge( v, w );

		// First check whether this graph was already created 
		if ( !knownGraphsComplete.contains( *evolved ) ) {

		    knownGraphsComplete.insert( *evolved );
		    
		    out->push( new Graph( *graph ) );

		    completeNodeFirstOrbit( v, evolved, out );
		} else {
		    delete evolved;
		}

	    }
	}
    }

    // Also consider the possibility that the current vertices won't have further edges.
    // attached to it.
    completeNodeFirstOrbit( v + 1, graph, out );
    
}

void GraphGenerator::complete( Graph* graph, ofstream* out ) {
    assert( graph->size() == totalVertices );

    // Flag vertices that have already minimal degree
    int nodesWithMaxDegree = 1; // Exclude root node.
    int nodesWithMinDegree = 1;	// - " -
    for ( int i = 1; i < totalVertices; ++i){
	int degree = graph->getDegree(i);
	if ( degree >= minDegree ) {
	    nodesWithMinDegree |= ( 1 << i );
	    if ( degree == maxDegree )
		nodesWithMaxDegree |= ( 1 << i );
	}
    }

    complete( graph, out, nodesWithMinDegree, nodesWithMaxDegree );
}

void GraphGenerator::complete( Graph* graph, ofstream* out, int minDegreeFlags, int maxDegreeFlags ) {
    // Search vertices (1 until totalVertices-1) for incomplete ones and complete those recursively.
    int v = orbit1;

    while( ++v < totalVertices ) {
	// Check if vertice already has minimal degree
	if ( minDegreeFlags & (1<<v) ) 
	    continue;	

	// Denots how many edges are needed until the vertex v
	// reaches minimal degree.
	int degreeLeft =  minDegree - graph->getDegree(v);

	// Use flags to prohibit edges to vertices that have already maximal degree or...
	unsigned int prohibited = maxDegreeFlags;
	// ...are already connected with the vertex...
	prohibited |= graph->A[v];
	// ...and the vertex itself.
	prohibited |= ( 1 << v );

	// This enumerates the different possibilities for a vertex to
	// have anonymous edges.
	for (int i = 0; i <= degreeLeft; ++i) {
	    completeNode( v, graph, prohibited, minDegreeFlags, maxDegreeFlags, (degreeLeft - i), i, out  );
	}
    }

    // When the Graph is complete put it in the results-list.
    // This condition will only work when all the bits in minDegreeFlags are set to one
    // i.e. all nodes have minimal degree.
    if ( minDegreeFlags + 1 == (1 << totalVertices) ) {

	graphBuffer.push_back( new Graph(*graph) );

	if ( graphBuffer.size() >= FLUSHING_LIMIT ) {
	    flushBuffer( out );   
	}

	return;
    }
}

void GraphGenerator::completeNode( int v, Graph* graph, int prohibited, int minDegreeFlags, int maxDegreeFlags, int degreeLeft, int anonymous, ofstream* out ){

    Graph* evolved = NULL ;

    // Mark current vertice as haveing minimal degree
    minDegreeFlags = minDegreeFlags | (1<<v);

    // If the degree equals 0 then only the anonymous edges have to be set.
    // This has to be done like this so that the division in 'getNextSubset'
    // does not cause a division by zero.
    if ( degreeLeft == 0 ){
	evolved = new Graph(*graph);
	evolved->setAnonymousEdges(v, anonymous);

	if ( !knownGraphsComplete.contains( *evolved )) {
	    knownGraphsComplete.insert( *evolved );	
	    complete( evolved, out, minDegreeFlags, maxDegreeFlags );
	}
    } else {
	// Otherwise we have to enumerate all possible connections within the graph.

	// endEnumeration is the subset with the biggest integer representation.
	int max = endEnumeration( degreeLeft, totalVertices );

	for ( unsigned int subset = beginEnumeration( degreeLeft ); subset <=  max; subset = getNextSubset(subset) ){
	    int newMinDegreeFlags = minDegreeFlags;
	    int newMaxDegreeFlags = maxDegreeFlags;
	
	    // This subsets tries to put edges which are not allowed.
	    // 'continue' gets the for-loop to take the next subset.
	    if ( (prohibited & subset) != 0)
		continue;

	    // Create new Graph, delete old.
	    if ( evolved != NULL ) 
		delete evolved;
	    
	    evolved = new Graph(*graph);

	    // Set anonymous edges.
	    evolved->setAnonymousEdges(v, anonymous);

	    // Set all edges given by the subset.
	    for(int i = 0; i < totalVertices ; ++i) {
		if ( (subset & (1 << i)) != 0 ) {

		    evolved->setEdge(v, i);
		    int degree = evolved->getDegree(i);

		    if ( degree >= minDegree ) {
			newMinDegreeFlags |= ( 1 << i );
			if ( degree == maxDegree )
			    newMaxDegreeFlags |= ( 1 << i );
		    }

		}
	    }

	    // Continue building the graph.
	    if ( !knownGraphsComplete.contains( *evolved ) ) {
		knownGraphsComplete.insert( *evolved );
		complete(evolved, out, newMinDegreeFlags, newMaxDegreeFlags );
	    }
	}

    }

    if ( evolved != NULL ) {
	delete evolved;
    }
}

void GraphGenerator::addPossibleEdges(Graph* graph, int v, ofstream* out ) {

    // Check whether graph is complete
    if ( v >= totalVertices ) {	
	// THOU SHALL NOT TEST WHETHER THIS GRAPH IS INSIDE KNOWNGRAPHSADDEDGES, FOR
	// IT WILL SMITE THE CORRECTNESS OF YOUR PROGRAM!
	graphBuffer.push( new Graph( *graph ) );	// create copy here so cleanup can be done as normal
	++graphCounter;
	delete graph;

	// Check if buffer should be flushed
	if ( graphBuffer.size() >= FLUSHING_LIMIT )
	    flushBuffer( out );


	return;
    }

    // Check whether further edges can be added, that is, the 
    // maximal Degree is not yet reached.
    if ( graph->getDegree(v) < maxDegree ) {
	// Try all possible edges (v,w) 
	for ( int w = 1; w < totalVertices; ++w ) { 
	    if ( v != w && (graph->getDegree(w) < maxDegree) ) {

		// Edge is possible, copy graph and add it. 
		Graph* evolved = new Graph( *graph );
		evolved->setEdge( v, w );

		// First check whether this graph was already created 
		if ( !knownGraphsAddEdges.contains( *evolved ) ) {

		    knownGraphsAddEdges.insert( *evolved );
		    graphBuffer.push( new Graph( *graph ) );
		    ++graphCounter;

		    // Check if buffer should be flushed
		    if ( graphBuffer.size() >= FLUSHING_LIMIT )
			flushBuffer( out );

		    addPossibleEdges( evolved, v, out );
		} else {
		    delete evolved;
		}

	    }
	}

	// Also add anonymous edge to vertices in orbit2.
	// As this happens AFTER the addition of further edges, the possibility of
	// |{anonymous Edges of v}| = 0 is taken into account.
	if ( v >= orbit1 + 1 ) {
	    Graph* evolved = new Graph( *graph );
	    evolved->setAnonymousEdges( v, evolved->getAnonymousEdges( v ) + 1 );

	    if ( !knownGraphsAddEdges.contains( *evolved ) ) {

		knownGraphsAddEdges.insert( *evolved );

		graphBuffer.push( new Graph( *evolved ) );
		++graphCounter;

		// Check if buffer should be flushed
		if ( graphBuffer.size() >= FLUSHING_LIMIT )
		    flushBuffer( out );

		addPossibleEdges( evolved, v, out );
	    } else {
		delete evolved;
	    }
	}

    }

    // Also consider the possibility that the current vertices won't have further edges.
    // attached to it.
    addPossibleEdges( graph, v + 1, out );
}

/*!
 *	Write all buffered graphs into the given output stream.
 */
void GraphGenerator::flushBuffer( ofstream* out ) {
    Graph* graph = NULL;
    while( graphBuffer.size() > 0 ) {
	graph = graphBuffer.peek();
	graphBuffer.pop();
	graph->encodeToFile( out );
	delete graph;

	bufferFlag = true;
    }

    out->flush();
}

int GraphGenerator::beginEnumeration(int elementNumber){

    int firstSubset = 0;
    for ( int i = 0; i < elementNumber; ++i){
	firstSubset |= (1 << i);
    }

    return firstSubset;

}

int GraphGenerator::endEnumeration(int elementNumber, int maxSize){

    int maxSet = 0;

    for ( int i = 0; i < elementNumber; ++i){
	maxSet |= ( 1 << (maxSize - i - 1) );
    }

    return maxSet;

}

unsigned int GraphGenerator::deleteFinishedBaseGraphs(List<Graph*>* baseGraphs) {

    int numberBaseGraphs = baseGraphs->size();
    
    if (minimalDegreeCompletionDone()){
	// Empty graphs without causing a leak.
	while( !baseGraphs->empty() ) {
	    delete baseGraphs->peek();
	    baseGraphs->pop();
	}
	return numberBaseGraphs + 1;
    }

    string completeGraphFilename = filename + "_completeGraphs";

    bufferFlag = false;	// We are using this flag to determine whether graphs where generated during this phase.

    // See if there exist graph files which we can use.
    int maxAlreadyDone = 0;
    for (maxAlreadyDone = 0; maxAlreadyDone < baseGraphs->size(); maxAlreadyDone++){
	stringstream sstream;
	sstream << completeGraphFilename  << maxAlreadyDone;
	if ( !fileExists( sstream.str() ))
	    break;
    }

    // Eliminate all the baseGraphs except the last one.
    for (int i = 0; i < maxAlreadyDone - 1; i++){
	delete baseGraphs->peek();
	baseGraphs->pop();
    }


    return maxAlreadyDone-1 < 0 ? 0 : maxAlreadyDone-1;
    
}

bool GraphGenerator::minimalDegreeCompletionDone() {

    string completeGraphFilename = filename + "_completeGraphsAllDone";

    if ( fileExists( completeGraphFilename ) )
	return true;

    return false;
    
}

void GraphGenerator::writeDummyFile() {

    stringstream sstream;
    string completeGraphFilename = filename + "_completeGraphs";
    
    sstream << completeGraphFilename << "AllDone";
    ofstream* out = new ofstream( sstream.str().c_str(), ios::binary );
    out->close();
    delete out;
    
}

string GraphGenerator::getCompleteFileName (int j) {

    stringstream sstream;
    string completeGraphFilename = filename + "_completeGraphs";

    sstream << completeGraphFilename << j;

    return sstream.str();
    
}

