#include "Global.h"
#include <time.h>
#include <cstring>
#include "../collection/ArrayPool.h"
#include "../HelpFunctions.h"

using namespace algebra;
using namespace collection;

int* algebra::berkowitzIterative( int** A, int n ) {
	int* res = ArrayPool::fetch( 2 );

	res[0] = 1;
	res[1] = -A[0][0];	

	for( int r = 1; r < n; ++r ) {
		int* T = calculateToeplitz( A, n, r );

		// Calculate T x rels
		int* temp = ArrayPool::fetch( r+2 );
		for( int i = 0; i <= r; ++i ) {
			temp[i] = 0;
			for( int j = 0; j <= i; ++j )
				temp[i] += res[j] * T[i-j];
		}
		temp[r+1] = 0;
		for( int j = 0; j < r+1; ++j )
			temp[r+1] += res[j] * T[r+1-j];

		ArrayPool::recycle( res, r+1 );
		ArrayPool::recycle( T, r+2 );

		res = temp;
	}

	return res;
}

int* algebra::calculateToeplitz( int** A, int n, int r ) {
	int* t = ArrayPool::fetch( r+2 );

	// First entry is always 1
	t[0] = 1;

	// Second entry is -A[r+1][r+1]
	t[1] = -A[r][r];

	// Create the leading r x r submatrix Ar of A and a second one to hold Ar^k
	// in the main loop.
	int** Ar = arrayInitializer( r, r );
	int** potAr = arrayInitializer( r, r );

	for ( int row = 0; row < r; ++row ){
		for ( int col = 0; col < r; ++col ){
			Ar[row][col] = Ar[col][row] = potAr[row][col] = potAr[col][row] = A[row][col];
		}
	}

	// Third entry is simply R x S
	t[2] = 0;
	for( int i = 0; i < r; ++i ) 
		t[2] -= A[r][i] * A[i][r];

	// Calculate other entries with R x Ar^(k-1) x S
	int* temp = ArrayPool::fetch( r ); 
	for( int k = 3; k < r+2; ++k ) {

		// Calculate current entry
		t[k] = 0;
		for ( int row = 0; row < r; ++row ) {
			int	inner = 0;
			for ( int col = 0; col < r; ++col ) 
				inner += A[r][col] * potAr[col][row];

			t[k] -= inner * A[row][r];
		}

		// Calculate Ar^k
		for( int row = 0; row < r; ++row ) {
			for( int col = 0; col < r; ++col ) {
				temp[col] = 0;
				for( int k = 0; k < r; ++k ) {
					temp[col] += potAr[row][k] * Ar[k][col];
				}
			}

			for ( int col = 0; col < r; ++col ){
				potAr[row][col] = temp[col];
			}
		}

	}
	ArrayPool::recycle( temp, r ); 

	// Clean up
	for ( int i = 0; i < r; ++i ){
		ArrayPool::recycle( Ar[i], r ); 
		ArrayPool::recycle( potAr[i], r ); 
	}
	delete[] Ar;
	delete[] potAr;

	return t;
}

int* algebra::berkowitz( int** A, unsigned int index, unsigned int dim ){   

	// Vector representing the Toeplitz-Matrix containing the coefficient for the next
	// recursive step.
	int width = dim-index;
	int matrixSize = width-1;
	int resultSize = dim+1-index;

	int* C = ArrayPool::fetch( resultSize ); // int* C = new int[dim+1-index];

	if ( resultSize == 2 ) {
		// stop recursion here
		C[0] = 1;
		C[1] = -A[0][0];
		return C;
	}

	/*
	 * Decompose Matrix A into the following sub-matrices ('a' being the top left entry only)
	 * | a    R |
	 * | S    M |
	 */
	int** M = arrayInitializer( matrixSize, matrixSize);
	int** potM = arrayInitializer( matrixSize, matrixSize);

	// fill M and potM, R and S. potM will be used to calculate M^(i-2) in the main loop.
	for ( int ix = 0; ix < matrixSize; ++ix ){
		for ( int iy = 0; iy < matrixSize; ++iy ){
			M[ix][iy] = potM[ix][iy] = A[ix+1][iy+1];
		}
	}

	// Calculate Toeplitz Matrix (only save first row as the rest can be simply reconstructed)
	// A. Handle special cases ( 0 <= i <= 2 )
	int entry;
	for ( int i = 0; i <= 2 && i < resultSize; ++i ){
		entry = 0;
		// trivial cases first
		if ( i == 0 ){
			// entries on the diagonal are 1 (use -1 b/c entry will be negated below)
			entry = -1;
		} else if ( i == 1 ) {
			// Entries below the diagonal are simply a (see above)
			entry = A[0][0];
		} else if ( i == 2 ) {
			// Special case: M^(2-2) = I, so the calculation can be simplified to the vector multiplication of R*S.
			for ( int i = 0; i < matrixSize; ++i ) {
				// entry += R[i] * S[i];
				entry += A[0][i+1] * A[i+1][0];
			}
		}
		C[i] = -entry;
	}

	// B. Handle regular case ( i >= 3 )
	for ( int i = 3; i < resultSize; ++i ) {
		entry = 0;

		// Calculate R * M^(i-2) * S
		for ( int ix = 0; ix < matrixSize; ++ix ) {
			int inner = 0;
			for ( int k = 0; k < matrixSize; ++k ) {
				// inner += R[k] * potM[ix][k];
				inner += A[0][k+1] * potM[ix][k];
			}
			// entry += inner * S[ix];
			entry += inner * A[ix+1][0];
		}

		// calculate next potM ( == M^(i-2+1) )
		int* temp = ArrayPool::fetch( matrixSize ); // int* temp = new int[matrixSize];            
		for( int row = 0; row < matrixSize; ++row ) {
			for( int col = 0; col < matrixSize; ++col ) {
				temp[col] = 0;
				for( int k = 0; k < matrixSize; ++k ) {
					temp[col] += potM[k][row] * M[col][k];
				}
			}

			for ( int col = 0; col < matrixSize; ++col ){
				potM[col][row] = temp[col];
			}
		}

		ArrayPool::recycle( temp, matrixSize ); // delete[] temp;

		C[i] = -entry;
	}

	// Clean up before recursion to save some memory
	for ( int i = 0; i < matrixSize; ++i ){
		ArrayPool::recycle( potM[i], matrixSize ); // delete[] potM[i];
	}

	delete[] potM;
	// ArrayPool::recycle( R, matrixSize ); // delete[] R;
	// ArrayPool::recycle( S, matrixSize ); // delete[] S;

	// Calculate result  
	int* nextN = berkowitz( M, index+1, dim ); 
	int* result = ArrayPool::fetch( resultSize ); // int* result = new int[resultSize];

	for ( int k = 0; k < resultSize; ++k ) {
		result[k] = 0;
		for ( int i = 0; i <= k && i < resultSize-1; ++i ) {
			result[k] += C[k-i] * nextN[i];
		}
	}

	ArrayPool::recycle( C, resultSize ); // delete[] C;

	// delete M...
	for ( int i = 0; i < matrixSize; ++i ){
		ArrayPool::recycle( M[i], matrixSize ); // delete[] M[i];
	}
	delete[] M;

	// nextN contains the result[] array of the next recursive step. Therefore its size is
	// dim+1-(index+1) == dim-index == resultSize-1
	ArrayPool::recycle( nextN, resultSize-1 ); // delete[] nextN;

	return result;
}

int** algebra::arrayInitializer(unsigned int width, unsigned int height){
	int** C = new int*[height];
	for ( int i = 0; i < height; ++i ){
		C[i] = ArrayPool::fetch( width ); // C[i] = new int[width];
		memset( C[i], 0, width*sizeof(int) );
	}

	return C;
}

int* algebra::charPoly( const Graph& G ) {
	static int maxCoeff = 0;
	static int minCoeff = 0;

	unsigned int dim = G.vertices;
	int* A = (int*)G.A;

	// Copy A into 2d integer array (assume A is symmetric and diagonal is 0)
	int **Araw = arrayInitializer( dim, dim );

	for ( int y = 0; y < dim; ++y ) {
		for ( int x = y; x < dim; ++x ) {
			Araw[x][y] = Araw[y][x] = (A[y] >> x) & 1;
		}
	}

	// Mark orbit2-nodes by writing a 1 on the diagonal-entries and
	// adding the number of anonymouse edges.
	Araw[0][0] = 1;
	for ( int i = G.orbit1 + 1; i < dim; ++i){
		Araw[i][i] = 1 + G.getAnonymousEdges(i);
	}

	// Calculate polynom 
	//int* result = berkowitz( Araw, 0, dim );
	int* result = berkowitzIterative(Araw, dim);

	/*
	// Max. coeff. for 4 7 was < 12000
	for (int i = 0; i < dim; i++) {
		if ( result[i] > maxCoeff ) {
			maxCoeff = result[i];
			PRINT( "Max. coeff. is now " << maxCoeff );
		}
		
		if ( result[i] < minCoeff ) {
			minCoeff = result[i];
			PRINT( "Min. coeff. is now " << minCoeff );
		}
	}
	*/

	// Clean up 
	for ( int x = 0; x < dim; ++x )
		ArrayPool::recycle( Araw[x], dim ); // delete[] Araw[x];

	delete[] Araw;

	return result;
}
