#include "FastGraphCollection.h"

using namespace collection;
using namespace std;
using namespace shared;
using namespace verifier;
using namespace help;

int** FastGraphCollection::permutationTables = NULL;

FastGraphCollection::FastGraphCollection() : graphs() {
	if ( permutationTables == NULL ) {
		int maxDegree = 7;
		permutationTables = new int*[maxDegree+1];

		for( int i = 1; i <= maxDegree; ++i ) {
			permutationTables[i] = generatePermutationTable( i );
		}
	}
}

FastGraphCollection::~FastGraphCollection() {
	for( List<Graph*>::iterator i = graphs.begin(); i != graphs.end(); ++i ) {
		delete *i;
	}

	graphs.clear();
}

unsigned int FastGraphCollection::size() const {
	return graphs.size();
}

/*
 *	Searches for a given graph and marks if it is found.
 *  @return true iff the graph could be found
 */
bool FastGraphCollection::flag( const Graph& g ) {
	// Asumption: graph is already unified.
	Graph* targetGraph = new Graph( g );

	int orbit1 = targetGraph->orbit1;
	for( List<Graph*>::iterator i = graphs.begin(); i != graphs.end(); ++i ) {
		Graph* currentGraph = *i;

		// Check simple equality (this is _not_ done in the following recursion )
		if ( *currentGraph == *targetGraph ) {
			currentGraph->flag = true;
			delete targetGraph;
			return true;
		} else {
			// Check if signatures are equal
			bool sig = true;
			for( int v = 0; v < targetGraph->vertices && sig; ++v ) {
				sig = targetGraph->getDegree( v ) == currentGraph->getDegree( v );
			}

			// Signatures differ: graphs cannot be isomorph
			if ( !sig )
				continue;

			// Create permutation (id)
			int* permutation = ArrayPool::fetch( targetGraph->vertices );

			for( int j = 0; j < targetGraph->vertices; ++j )
				permutation[j] = j;

			// Determine first degree-group
			int firstGroup = 1;
			int degree = targetGraph->getDegree(1);	// start with first vertex from orbit1 
			while( targetGraph->getDegree(++firstGroup) == degree && firstGroup <= orbit1 );

			bool isomorph = compare( targetGraph, currentGraph, permutation, 1, firstGroup );		

			ArrayPool::recycle( permutation, targetGraph->vertices );

			if ( isomorph ) {
				// Found isomorph graph
				currentGraph->flag = true;
				delete targetGraph;
				return true;			
			}
		}
	}	
	delete targetGraph;
	return false;
}

/*
 *  Inserts a given graph into this collection.
 */
bool FastGraphCollection::insert( const Graph& g, bool deleteOne ) {
	// Assumption: graph is unified (see GraphTrie).
	Graph *h = new Graph( g );

	bool deleted = false;

	if (deleteOne){
		if ( !graphs.empty() ){
			graphs.pop();
			deleted = true;
		}
	}

	graphs.push_back( h );

	return deleted || !deleteOne;
}

/*
 *	Checks whether a given graph is contained in this collection.
 *	@return true iff an isomorph graph was found
 */
bool FastGraphCollection::contains( const Graph& g ) {
	// Assumption: graph is unified (see GraphTrie).
	Graph* targetGraph = new Graph( g );

	int orbit1 = targetGraph->orbit1;
	bool contained = false;
	for( List<Graph*>::iterator i = graphs.begin(); i != graphs.end() && !contained; ++i ) {
		Graph* currentGraph = *i;

		// Check simple equality (this is _not_ done in the following recursion )
		if ( *currentGraph == *targetGraph ) {
			contained = true;		
		} else {
			// Check if signatures are equal
			bool sig = true;
			for( int v = 0; v < targetGraph->vertices && sig; ++v ) {
				sig = targetGraph->getDegree( v ) == currentGraph->getDegree( v );
			}

			// Signatures diff, graphs cannot be isomorph
			if ( !sig )
				continue;

			// Create permutation (id)
			int* permutation = ArrayPool::fetch( targetGraph->vertices );

			for( int j = 0; j < targetGraph->vertices; ++j )
				permutation[j] = j;

			// Determine first degree-group
			int firstGroup = 1;
			int degree = targetGraph->getDegree(1);	// start with first vertex from orbit1 
			while( targetGraph->getDegree(++firstGroup) == degree && firstGroup <= orbit1 );

			contained = compare( targetGraph, currentGraph, permutation, 1, firstGroup );		

			ArrayPool::recycle( permutation, targetGraph->vertices );
		}
	}
	delete targetGraph;
	return contained;
}

/*
 *	Checks whether targetGraph and currentGraph are isomorph under the permutation p.
 *	All vertices with indices smaller than groupStart are isomorph; all vertices with indices in
 *  [groupdStart, groupEnd) are checked during this call and all vertices with indices greater or equal
 *  groupEnd will be checked recursively.
 *
 *  Vertices are checked group-wise, that is, vertices with the same degree are permuted inside the targetGraph 
 *  (permutation which map vertices with different degree onto each other can be discarded). As the graph is unified
 *  (the vertices are sorted by their degree inside each orbit), vertex groups can be identified by a start and an end index.
 *  Groups do not overlap the orbit boundaries, as we assume that valid permutations only map same-orbit nodes onto each other,
 *  due to the unknown nature of the anonymous edges of orbit 2.
 *
 *  If the currently examined group of same-degree vertices in 'targetGraph' is found to be equal to the corresponding group inside 'currentGraph',
 *  the next group is determined and checked recursively in the same fashion.
 */
bool FastGraphCollection::compare( Graph* targetGraph, Graph* currentGraph, int* p, int groupStart, int groupEnd ) {

	int n = targetGraph->vertices;
	int orbit1 = targetGraph->orbit1;

	// Determine next equal-degree group of vertices.
	int nextGroupStart = groupEnd, nextGroupEnd = groupEnd;
	if ( nextGroupStart <= orbit1 ) {
		// Group is still in orbit1
		int degree = targetGraph->getDegree( nextGroupEnd );
		while( ++nextGroupEnd <= orbit1 && targetGraph->getDegree( nextGroupEnd ) == degree );
	} else if ( nextGroupEnd < n ) {
		// Group is in orbit2 and groupEnd is still a valid vertex
		int degree = targetGraph->getDegree( nextGroupEnd );
		while( ++nextGroupEnd < n && targetGraph->getDegree( nextGroupEnd ) == degree );
	}

	if ( groupStart < n ) {
		// Try all permutations of vertices in {groupStart, ..., groupEnd-1}.
		// The permutation table indicates which elements have to be swapped to
		// succesively generate all possible permutations.
		int* swaps = getPermutationTable( groupEnd-groupStart );

		for( int i = 1; i < swaps[0]; ++i ) {
			// Apply current swap to permutation: swap swaps[i] with swaps[i]-1
			int temp = p[ groupStart+swaps[i] ];
			p[ groupStart+swaps[i] ] = p[ groupStart+swaps[i]-1 ];
			p[ groupStart+swaps[i]-1 ] = temp;

			// First check whether the graphs are equal under the current permutation
			// up to 'groupEnd'. If that is the case, continue working on this permutation. 
			if ( checkPermutation( targetGraph, currentGraph, p, groupEnd ) 
					&& compare( targetGraph, currentGraph, p, nextGroupStart, nextGroupEnd ) )
				return true;
		}

		if ( groupEnd-groupStart > 1 ) {
			// Revert permutation to original configuration (one last swap)
			int temp = p[groupStart];
			p[groupStart] = p[groupStart+1];
			p[groupStart+1] = temp;
		}

		// check this permutation (it was _not_ checked in the loop above)
		if ( checkPermutation( targetGraph, currentGraph, p, groupEnd ) 
				&& compare( targetGraph, currentGraph, p, nextGroupStart, nextGroupEnd ) )
			return true;
	} else {
		// Check whether graphs are equal under this permutation
		return checkPermutation( targetGraph, currentGraph, p, n );
	}

	return false;
}

bool FastGraphCollection::checkPermutation( Graph* targetGraph, Graph* currentGraph, int* p, int limit ) {
	// DEBUG:
	/*
	   cout << "      ";
	   for( int i = 0; i < limit; ++i ) {
	   cout << p[i] << ", ";
	   }	
	   cout << endl;
	   */

	bool permOkay = true;
	for( int v = 1; v < limit && permOkay; ++v ) {
	  permOkay = currentGraph->getAnonymousEdges(v) == targetGraph->anonymousEdges[ p[v] ];
		for( int w = v+1; w < limit && permOkay; ++w ) {
			permOkay = currentGraph->adjacent( v,w ) == targetGraph->adjacent( p[v], p[w] );
		}
	}

	return permOkay;
}

/*
 *	Generates a series of swaps that, if applied sucessively to an array,
 *  result in an enumeration of all permutations. After the last swap is applied
 *	all elements except the first two are again at their respective place (swapping
 *	the first two elements therefore results in the initial configuration).
 *
 *	The nth element of this array t[] means that element t[n]-1 and t[n] are swapped.
 *
 *	It is a sin to disbelieve Knuth's holy word. Hail Knuth.
 */
int* FastGraphCollection::generatePermutationTable( int n ) {
	int fac = n;
	for( int i = n-1; i > 0; --i )
		fac *= i;

	// See TAOCP fasc 2b
	// T1
	int* t = new int[ fac ];
	int N = fac;
	int d = N/2;
	int m = 2;
	int k = 0;
	t[ d ] = 1;

	// T2
	while( m < n ) {
		++m;
		d = d/m;
		k = 0;

		do {
			// T3
			k += d;
			int j = m-1;
			while( j > 0 ) {
				t[k] = j;
				k += d;
				--j;
			}

			// T4
			++t[k];
			k += d;

			// T5
			while( j < m-1 ) {
				++j;
				t[k] = j;
				k += d;			
			}
		} while( k < N );
	}

	// write length of permutation into t[0]
	t[ 0 ] = fac;
	return t;
}

int* FastGraphCollection::getPermutationTable( int n ) {
	return permutationTables[n];
}

bool FastGraphCollection::flagGraph( Graph* g ) {
	for( List<Graph*>::iterator i = graphs.begin(); i != graphs.end(); ++i ) {
		if ( *g == *(*i) ) {
			(*i)->flag = true;
			return true;
		}
	}
	return false;
}

void FastGraphCollection::clear() {
	graphs.clear();
}

int FastGraphCollection::allFlagged(List<Graph*>* notFlagged) {
	int res = 0;

	for( List<Graph*>::iterator i = graphs.begin(); i != graphs.end(); ++i ) {
		if( !(*i)->flag ) {
			notFlagged->push_back(*i);
			++res;
		}
	}

	return res;
}

bool FastGraphCollection::empty() {
	return graphs.empty();
}

void FastGraphCollection::pop() {
	graphs.pop();
}

Graph* FastGraphCollection::peek() {
	return graphs.peek();
}

/*
   void FastGraphCollection::unify( Graph* g ) {
// Unify the graph representation by sorting vertices inside their respective orbits
// by degree.
int orbit1 = g->orbit1;	
int orbit2 = g->orbit2;	

int* degree = ArrayPool::fetch( 1+orbit1+orbit2 );
for( int i = 0; i < 1+orbit1+orbit2; ++i ) {
degree[ i ] = g->getDegree( i );
}

// Sort orbit1
for( int i = 1; i <= orbit1; ++i ) {
int min = i;

for( int k = i+1; k <= orbit1; ++k ) 
if( degree[ k ] < degree[ min ] ) min = k;
if ( i == min ) continue;

// swap
int temp = degree[i];
degree[i] = degree[min];
degree[min] = temp;

g->swapVertices( i, min );
}	


// Sort orbit2
for( int i = orbit1+1; i <= orbit1+orbit2; ++i ) {
int min = i;

for( int k = i+1; k <= orbit1+orbit2; ++k ) 
if( degree[ k ] < degree[ min ] ) min = k;
if ( i == min ) continue;

// swap
int temp = degree[i];
degree[i] = degree[min];
degree[min] = temp;

g->swapVertices( i, min );
}

ArrayPool::recycle( degree, 1+orbit1+orbit2 );
}
*/
