#include "GraphTrie.h"
#include <cstring>

using namespace collection;
using namespace verifier;
using namespace std;
using namespace algebra;

/*
 *	Creates a trie that stores graphs with orbits inside the given boundaries.
 */
GraphTrie::GraphTrie( unsigned int minOrbit1, unsigned int maxOrbit1, unsigned int minOrbit2, unsigned int maxOrbit2,
		unsigned int maxNumberEntries) {
	root = new TrieNode;
	height = graphs = 0;

	this->maxNumberEntries = maxNumberEntries;

	checkAll = true;
}

GraphTrie::~GraphTrie() {
	// delete all nodes recursivly
	clear( root );
}

/*
 * Calculates a characteristic polynom of the given graph. The contract here is that the polynomials
 * of two isomorphic graphs are equal.
 * @param g the graph for which the polynomial should be calculated
 * @param polySize a pointer to an integer into which the size of the polynomial will be written.
 *
 * @return the coefficients of the polynomial as an array of integers (index equals degree of term)
 */
int* GraphTrie::calculatePolynom( const Graph& g, int* polySize ) {
	// Memorize last graph with result
	static Graph lastGraph( 1, 1 );
	static int lastPolynomSize = 999999;	// See ArrayPool::recycle
	static int* lastPolynom = new int[1];

	// Check if result really needs to be calculated again
	int *C;
	if ( g == lastGraph && lastPolynom != NULL ) {
		C = lastPolynom;	// Use polynom from last time
	} else {
		ArrayPool::recycle( lastPolynom, lastPolynomSize );	// Delete old polynom 
		C = algebra::charPoly( g );	// Reacalculate polynom
	}

	// The first coefficient should always be '1'. It is therefore excluded in the trie.
	assert( C[0] == 1 );

	// Size of the polynomial is g.vertices, but we can exclude the first coefficient.
	int polynomSize = g.vertices-1;

	// Hand back the size of the polynomial via the given pointer.
	*polySize = polynomSize;

	// Create a copy of polynome excluding the first element.
	int* p = ArrayPool::fetch( polynomSize );
	memcpy( p, C+1, polynomSize*sizeof(int) );

	// Memorize values
	lastGraph = g;
	lastPolynom = C;
	lastPolynomSize = g.vertices;

	return p;
}

/*
 * Inserts a given graph into the trie.
 */
void GraphTrie::insert( const Graph& g ) {

	// Unify graph (sort vertices by degree inside their respective orbits)
	Graph* unified = new Graph( g );
	unified->unify();

	// Calculate characteristic polynomial.
	int polySize = 0;
	int* charPoly = calculatePolynom( *unified, &polySize );

	// Insert graph into appropriate node.
	insert( *unified, root, charPoly, polySize, 0 );

	// Delete polynomial and graph.
	ArrayPool::recycle( charPoly, polySize );
	delete unified;
}

/*
 *  Checks whether a given graph (or an isomorphic version) is contained in this trie.
 */
bool GraphTrie::contains( const Graph& g ) {
	// Unify graph (sort vertices by degree inside their respective orbits)
	Graph* unified = unify( new Graph( g ) );

	// Calculate characteristic polynomial.
	int polySize = 0;
	int* charPoly = calculatePolynom( *unified, &polySize );

	// Checks whether the graph is contained in this trie.
	bool res = contains( *unified, root, charPoly, polySize );

	// Delete polynomial and graph.
	ArrayPool::recycle( charPoly, polySize );
	delete unified;

	return res;
}

/*
 *  Checks whether a given graph (or an isomorphic version) is contained in this trie and flags it.
 */
bool GraphTrie::flag( const Graph& g ) {
	// Unify graph (sort vertices by degree inside their respective orbits)
	Graph* unified = unify( new Graph( g ) );

	// Calculate characteristic polynomial.
	int polySize = 0;
	int* charPoly = calculatePolynom( *unified, &polySize );

	// Checks whether the graph is contained in this trie and flags it.
	bool res = flag( *unified, root, charPoly, polySize );

	// Delete polynomial and graph.
	ArrayPool::recycle( charPoly, polySize );
	delete unified;

	return res;
}

/*
 *  Recursively descends the trie until the right node to insert this graph is found.
 *  termIndex denotes the currently examined coefficient of the polynomial.
 */
void GraphTrie::insert( const Graph& g, TrieNode* currentTrieNode, int* poly, int termIndex, unsigned int depth ) {
	// The term of the polynom that is used at this step to know where to
	// insert the graph is read.
	int currentCoefficient = (poly[ --termIndex ]);

	// Keep track of the height
	height = depth > height ? depth : height;

	TrieEntry* currentTrieEntry = NULL;

	// Finds the position of the current coefficient inside the current node or, if the coefficien is not found, the right
	// position to insert it. 
	unsigned long position = findCoefficient( currentTrieNode, currentCoefficient );

	// Check whether match was found, otherwise 'position' indicates where to insert the new coefficient.
	// 'position' might equal currentTrieNode->size, meaning the new entry must be appended to the current array.
	if ( position < currentTrieNode->size && currentTrieNode->entries[position]->coefficient == currentCoefficient ) {	
		currentTrieEntry = currentTrieNode->entries[position];
	}

	// If this is the case, the polynom of the graph has some term which did not appear
	// before so a new entry has to be generated.
	if ( currentTrieEntry == NULL ) {
		// Create the new trie entry with the current coefficient.
		currentTrieEntry = new TrieEntry( currentCoefficient );
		currentTrieEntry->child = new TrieNode();

		// Insert entry into current node by enlarging the array by one.
		currentTrieNode->size++;
		TrieEntry** entries = new TrieEntry*[ currentTrieNode->size ];

		// Copy first part of the entry array.
		memcpy( entries, currentTrieNode->entries, sizeof(TrieEntry*) * position );

		// Insert the new element.
		entries[ position ] = currentTrieEntry;

		// Copy the second part of the array.
		memcpy( entries+(position+1), 
				currentTrieNode->entries+position, 
				sizeof(TrieEntry*)*(currentTrieNode->size-position-1) );

		// Delete old array and replace with new one
		if ( currentTrieNode->entries != NULL )
			delete[] currentTrieNode->entries;
		currentTrieNode->entries = entries;

	}

	// Check whether last term was reached, if that is the case, the graph is inserted at the current trie-entry.
	// Otherwise the next coefficient must be used to further descend the trie.
	if ( termIndex == 0 )
		insertAt( g, currentTrieEntry );
	else
		insert( g, currentTrieEntry->child, poly, termIndex, depth+1 );
}

/*
 *  Inserts a graph into the given trie entry and keeps track of the number of graphs.
 */
void GraphTrie::insertAt( const Graph& g, TrieEntry* entry ) {
	if ( entry->graphs == NULL )
		entry->graphs = new FastGraphCollection();

	bool deleteOne = maxNumberEntries != 0 && graphs >= maxNumberEntries; 

	bool deleted = entry->graphs->insert( g, deleteOne );

	if ( !deleted )
		this->deleteOneGraph(root);

	graphs += graphs >= maxNumberEntries ? 0 : 1;
}

bool GraphTrie::deleteOneGraph(TrieNode* currentTrieNode) {

	for ( int i = 0; i < currentTrieNode->size; ++i ) {
		TrieEntry* currentTrieEntry = currentTrieNode->entries[i];
		if ( currentTrieEntry->graphs != NULL && !currentTrieEntry->graphs->empty() ){
			currentTrieEntry->graphs->pop();
			return true;
		}

		if ( deleteOneGraph( currentTrieEntry->child ) ){
			return true;
		}

	}
	return false;
}


/*
 * Checks whether the given trie-entry contains a given graph.
 */
bool GraphTrie::containsGraph( const Graph& g, TrieEntry* entry ) {
	return entry->graphs != NULL && entry->graphs->contains( g );
}

/*
 *	Flags a given graph inside a trie-entry and returns whether it was found.
 */
bool GraphTrie::flagGraph( const Graph& g, TrieEntry* entry ) {
	return entry->graphs != NULL && entry->graphs->flag( g );
}

/*
 * Checks recursively if a given graph is contained in the trie.
 */
bool GraphTrie::contains( const Graph& g, TrieNode* currentTrieNode, int* poly, int termIndex ) {
	int currentCoefficient = (poly[ --termIndex ]);
	bool lastRow = termIndex == 0;

	// Finds the position of the current coefficient inside the current node.
	unsigned long position = findCoefficient( currentTrieNode, currentCoefficient );

	TrieEntry* currentTrieEntry;

	// If the current coefficient is larger than all coefficients contained
	// in this trie-node, 'position' equals currentTrieNode->size+1 and the search
	// can be aborted.
	if ( position >= currentTrieNode->size ) 
		return false;

	// Continue searching; if the coefficient at position does not match the search coefficient
	// the latter is not contained in the current trie-node; the search can be aborted.
	if ( currentTrieNode->entries[position]->coefficient == currentCoefficient ) {	
		currentTrieEntry = currentTrieNode->entries[position];
		if ( lastRow )
			return containsGraph( g, currentTrieEntry );
		else
			return contains( g, currentTrieEntry->child, poly, termIndex );
	}

	return false;
}

/*
 * Recursively searches for a given graph and flags it if it was found (also considers
 * isomorphic versions).
 */
bool GraphTrie::flag( const Graph& g, TrieNode* currentTrieNode, int* poly, int termIndex ) {
	int currentCoefficient = (poly[ --termIndex ]);
	bool lastRow = termIndex == 0;

	unsigned long position = findCoefficient( currentTrieNode, currentCoefficient );

	assert( position >= 0 );
	assert( position <= currentTrieNode->size );

	TrieEntry* currentTrieEntry;

	// Check whether match was found
	if ( position >= currentTrieNode->size ) 
		return false;

	if ( currentTrieNode->entries[position]->coefficient == currentCoefficient ) {	
		currentTrieEntry = currentTrieNode->entries[position];
		if ( lastRow )
			return flagGraph( g, currentTrieEntry );
		else
			return flag( g, currentTrieEntry->child, poly, termIndex );
	}

	return false;
}

void GraphTrie::clear() {
	clear( root );
	root = new TrieNode;    // create new root
	height = graphs = 0;
}

/*
 *  Empties this trie.
 */
void GraphTrie::clear( TrieNode* currentTrieNode ) {
	for ( int i = 0; i < currentTrieNode->size; ++i ) {
		clear( currentTrieNode->entries[i]->child );
		delete currentTrieNode->entries[i];
	}

	delete[] currentTrieNode->entries;
	delete currentTrieNode;
}


unsigned int GraphTrie::size() {
	return graphs;
}

unsigned int GraphTrie::getHeight() const {
	return height;
}

int GraphTrie::allFlagged(ofstream* file){

	List<Graph*>* notFlagged = new List<Graph*>();

	int res = allFlagged(this->root, notFlagged);

	if ( file != NULL ) {
		for ( List<Graph*>::iterator i = notFlagged->begin(); i != notFlagged->end(); i++ ) {
			(*i)->encodeToFile(file);
		}
	}

	return res;
}

/* 
 *  Checks whether all graphs inside this trie have been flagged. Assuming that each graph is
 *  contained exactly once (counting isomorphic versions) in this trie, this method can be used to assert that two sets of graphs
 *  are equal (in the one direction by checking whether every graph is contained, in the  other by checking
 *  whether all graphs have been flagged).
 */
int GraphTrie::allFlagged( TrieNode* currentTrieNode, List<Graph*>* notFlagged ) {
	int result = 0;

	for ( int i = 0; i < currentTrieNode->size; ++i ) {
		TrieEntry* currentTrieEntry = currentTrieNode->entries[i];
		if ( currentTrieEntry->graphs != NULL )
			result += currentTrieEntry->graphs->allFlagged(notFlagged);
		result += allFlagged( currentTrieEntry->child, notFlagged );
	}

	return result;
}

/* Unifies the graph representation by sorting the vertices inside their respective orbits
 * by degree (ascending).
 */
Graph* GraphTrie::unify( Graph* g ) {
	int orbit1 = g->orbit1;	
	int orbit2 = g->orbit2;	

	// Store degree in separate array to avoid calling getDegree() every time.
	int* degree = ArrayPool::fetch( 1+orbit1+orbit2 );
	for( int i = 0; i < 1+orbit1+orbit2; ++i ) {
		degree[ i ] = g->getDegree( i );
	}

	// Sort orbit1 (selection sort)
	for( int i = 1; i <= orbit1; ++i ) {
		int min = i;

		// Find minimal element inside (i, orbit1]
		for( int k = i+1; k <= orbit1; ++k ) 
			if( degree[ k ] < degree[ min ] ) min = k;
		if ( i == min ) continue;

		// Swap degrees
		int temp = degree[i];
		degree[i] = degree[min];
		degree[min] = temp;

		// Swap vertices
		g->swapVertices( i, min );
	}	


	// Sort orbit2 (selection sort)
	for( int i = orbit1+1; i <= orbit1+orbit2; ++i ) {
		int min = i;

		// Find minimal element inside (i, orbit1+orbit2]
		for( int k = i+1; k <= orbit1+orbit2; ++k ) 
			if( degree[ k ] < degree[ min ] ) min = k;
		if ( i == min ) continue;

		// Swap degrees
		int temp = degree[i];
		degree[i] = degree[min];
		degree[min] = temp;

		// Swap vertices
		g->swapVertices( i, min );
	}

	ArrayPool::recycle( degree, 1+orbit1+orbit2 );
	return g; 
}

/*
 *	Conducts a binary search on the current trie node's entries. It returns either the position
 *	of the desired coefficient or the position where that coefficient should be inserted (therefore
 *	a check is required whether the searched element actually is at the resulting position).
 */
unsigned long GraphTrie::findCoefficient( TrieNode* currentTrieNode, int coefficient ) {
	if ( currentTrieNode->size == 0 ) return 0;

	// Binary search the coefficient.
	unsigned long low = 0;
	unsigned long high = currentTrieNode->size;
	while( low < high ) {
		unsigned long mid = (low+high)/2;
		if ( currentTrieNode->entries[mid]->coefficient < coefficient ) 
			low = mid + 1;
		else
			high = mid;
	}

	// This is either the right position or the position where the coefficient
	// should be inserted.
	return low;
}

/*
 * For debugging
 */
/*
   void GraphTrie::sizeNodes(){
   sizeNodes(this->root);
   }

   void GraphTrie::sizeNodes(TrieNode* currentTrieNode){

   for ( int i = 0; i < currentTrieNode->size; ++i ) {
   int size = currentTrieNode->entries[i]->graphs-.vertices;
   if ( size != 0 )
   std::cout << size  << ", ";
   }

   for ( int i = 0; i < currentTrieNode->size; ++i ) {
   sizeNodes( currentTrieNode->entries[i]->child );
   }

   }

   void GraphTrie::print() {
   cout << " size  | height | nodes " << endl;
   cout << "  " << graphs << "   " << height << "   " << entries << endl << flush;
   }
   */
