#ifndef _LIST_H_
#define _LIST_H_

#include <assert.h>

namespace collection {

   /*
    * An element in the list is represented by one of
    * these ListNodes.
    */ 
   template<class T>
      struct ListNode {
	 T element;
	 ListNode* next;
      ListNode( const T& element ) : element( element ) {
	 next = 0;
      }
      };

   /*
    * A simple linked list.
    */
   template<class T>
      class List {
   private:
      // First element.
      ListNode<T>* first;

      // Last element
      ListNode<T>* last;

      // A counter for the number of elements in the list.
      unsigned int counter;

   public:
      /*
       * Instantiates an empty list.
       */
      List() {
	 counter = 0;
	 first = last = 0;
      }

      ~List() {
	 ListNode<T>* current = first;
	 ListNode<T>* last;
	 while( current != 0 ) {
	    last = current;
	    current = current->next;
	    delete last;
	 }
      }

      /*
       * Inserts an element to the the list at the first position.
       *
       * @return a reference to this list for chaining.
       */
      List& push( const T& element ) {
	 ListNode<T>* temp = first;
	 first = new ListNode<T>( element );
	 first->next = temp;
	 ++counter;

	 // Necessary for the first element, since 'last' and 'first'
	 // must be then the same element.
	 if( last == 0 ) last = first;

	 return *this;
      }

      /*
       * Inserts an element at the back of the list.
       *
       * @return a reference to this list for chaining.
       */
      List& push_back( const T& element ) {
	 if ( first != 0 ) {
	    assert( last != 0 );
	    last->next = new ListNode<T>( element ); // WTF
	    last = last->next;
	    last->next = 0;
	 } else { // The list is empty.
	    assert( last == 0 );
	    first = new ListNode<T>( element );
	    first->next= 0;
	    last = first;
	 }
	 ++counter;

	 return *this;
      } 

      /*
       * Deletes the first element in the List. Fails if the list is empty.
       *
       * @return a reference to this list for chaining.
       */
      List& pop() {
	 assert( first != 0 );
	 ListNode<T>* temp = first;
	 first = first->next;
	 delete temp;

	 --counter;

	 //Check if the List is now empty.
	 if( first == 0 ) last = 0;

	 return *this;
      }

      /*
       * Returns a pointer to the first element in the list. Fails if the list is empty.
       *
       * @return a reference to this list for chaining.
       */
      T& peek() const {
	 assert( first != 0 );
	 return first->element;
      }

      /*
       * Empties the List.
       */
      void clear() {
	 ListNode<T>* current = first;
	 ListNode<T>* previous;
	 while( current != 0 ) {
	    previous = current;
	    current = current->next;
	    delete previous;
	 }

	 first = last = 0;
	 counter = 0;
      }

      /*
       * Returns the size of the List, i.e. how many elements the
       * List contains.
       */
      unsigned int size() const {
	 return counter;
      }

      /*
       * Returns true iff the List is empty.
       */ 
      bool empty() const {
	 return first == 0;
      }


      /*
       * An iterator for the List.
       */
      class iterator {
      private:
	 friend class List<T>;
	 ListNode<T>* current;
      public:
	 iterator() {
	    current = 0;
	 }

	 iterator( const iterator& other ) {
	    current = other.current;
	 }

	 T& operator*() {
	    return current->element;
	 }

	 iterator& operator=( const iterator& other ) {
	    current = other.current;
	    return *this;
	 }

	 iterator& operator++() {
	    assert( current != 0 );
	    current = current->next;
	    return *this;
	 }

	 iterator& operator++(int) {
	    assert( current != 0 );
	    current = current->next;
	    return *this;
	 }

	 bool operator==( const iterator& other ) {
	    return current == other.current;
	 }

	 bool operator!=( const iterator& other ) {
	    return current != other.current;
	 }

      };

      iterator begin() {
	 iterator i;
	 i.current = first;
	 return i;
      }

      iterator end() {
	 iterator i;
	 i.current = 0;
	 return i;
      }


   };


}

#endif
