#include "BitMatrix.h"

#include <iostream>
#include <assert.h>

using namespace shared;
using namespace std;

BitMatrix::BitMatrix( unsigned int w, unsigned int h ) : width(w), height(h) {
    static bool warned = true;
    if ( !warned ) {
        warned = true;
        cout << "using bitmatrix" << endl;
    }
    
    rows = new BitVector[ height ];
    for ( unsigned int i = 0; i < height; ++i )
        rows[ i ] = BitVector( width );
    
}

BitMatrix::BitMatrix( const BitMatrix& other ) {
    height = other.height;
    width = other.width;
    
    rows = new BitVector[ height ];
    for ( unsigned int i = 0; i < height; ++i )
        rows[ i ] = BitVector( other.rows[ i ] );
}

BitMatrix::BitMatrix( const BitMatrix* other ) {
    height = other->height;
    width = other->width;
    
    rows = new BitVector[ height ];
    for ( unsigned int i = 0; i < height; ++i )
        rows[ i ] = BitVector( other->rows[ i ] );
}

BitMatrix::~BitMatrix() {
    if ( rows ) {
        delete[] rows;
    }
}

unsigned int BitMatrix::getWidth() const {
    return width;
}

unsigned int BitMatrix::getHeight() const {
    return height;
}


/*
 *  BitMatrix modification methods
 */

BitMatrix BitMatrix::cut( unsigned int x, unsigned int y, unsigned int w, unsigned int h ) {
    assert( x >= 0 && x < width && y >= 0 && y < height && w > 0 && h > 0);
    assert( x+w <= width && y+h <= height );
    
    BitMatrix res( w, h );
    
    for ( unsigned int iy = 0; iy < h; ++iy )
        res.rows[ iy ] = (rows[ iy+y ]).cut( x, w );
    
    return res;
}

BitMatrix& BitMatrix::paste( unsigned int x, unsigned int y, const BitMatrix& other ) {
    assert( x >= 0 && x < width && y >= 0 && y < height );
    assert( x+other.width <= width && y+other.height <= height );
    
    for ( unsigned int iy = 0; iy < other.height; ++iy ) {
        (rows[ iy+y ]).paste( x, other.rows[ iy ]);
    }
    
    
    return *this;
}

BitMatrix& BitMatrix::set( unsigned int x, unsigned int y ) {
    assert( x < width && y  <height && x >= 0 && y >= 0 );
    
    rows[ y ].set( x );
    
    return *this;
}


BitMatrix& BitMatrix::unset( unsigned int x, unsigned int y ) {
    assert( x < width && y  <height && x >= 0 && y >= 0 );
    
    rows[ y ].unset( x );
    
    return *this;
}

bool BitMatrix::isSet( unsigned int x, unsigned int y ) const {
    assert( x < width && y  <height && x >= 0 && y >= 0 );
    
    return rows[ y ].isSet( x );
}

BitMatrix& BitMatrix::clear( bool value ) {
    for ( unsigned int i = 0; i < height; ++i )
        rows[ i ].clear( value );
    
    return *this;
}


BitVector BitMatrix::getRow( unsigned int y ) const {
    assert( y < height );
    
    BitVector res( rows[ y ] );
    return res;
}

void BitMatrix::setRow( unsigned int y, const BitVector& row ) {
    assert( y < height );
    assert( row.size() == width );
    
    rows[ y ] = row;
}

void BitMatrix::removeRow( unsigned int y ) {
    assert( y < height );
    
    BitVector* oldRows = rows;
    rows = new BitVector[ height-1 ];
    
    for ( int i = 0; i < y; ++i )
        rows[ i ] = oldRows[ i ];
    
    for( int i = y+1; i < height; ++i )
        rows[ i-1 ] = oldRows[ i ];
    
    delete[] oldRows;
    
    height--;
}

BitVector BitMatrix::getColumn( unsigned int x ) const {
    assert( x < width );
    
    BitVector res( height );
    for ( int y = 0; y < height; ++y ) {
        if ( rows[ y ].isSet( x ) ) res.set( y );
    }
    
    return res;
}

void BitMatrix::setColumn( unsigned int x, const BitVector& v ) {
    assert( x < width );
    assert( v.size() == height );
    
    for( int y = 0; y < height; ++y ) {
        if ( v.isSet( y ) )
            rows[ y ].set( x );
        else
            rows[ y ].unset( x );
    }
}

void BitMatrix::removeColumn( unsigned int x ) {
    assert( x < width );
    
    if ( x == width-1 ) {
        // just erase last bit
        for ( int y = 0; y < height; ++y ) {
            rows[ y ] = rows[ y ].cut( 0, width-1 );
        }
    } else {
        for ( int y = 0; y < height; ++y ) {
            BitVector& oldRow = rows[ y ];
            oldRow.paste( x, oldRow.cut( x+1, width-x-1  ) );
            rows[ y ] =  oldRow.cut( 0, width-1 );
        }
    }
    
    width--;
}

void BitMatrix::removeRowAndColumn( unsigned int i ) {
    assert( i < width && i < height );
    removeRow( i );
    removeColumn( i );
}

void BitMatrix::swapRowsAndColumns( unsigned int a, unsigned int b ) {
    /*
     * This should eventually be replaced by a specialized method for symmetric matrices
     */
    assert( a < width && b < width && a < height && b < height );
    
    if ( a == b ) return;	// this makes sense since permutations encoded in arrays will be using this
    
    BitVector rowA = getRow( a );
    BitVector rowB = getRow( b );
    BitVector colA = getColumn( a );
    BitVector colB = getColumn( b );
    
    // save crossing bits
    bool aa = rowA.isSet( a );
    bool ab = rowA.isSet( b );
    bool ba = rowB.isSet( a );
    bool bb = rowB.isSet( b );
    
    setRow( b, rowA );
    setRow( a, rowB );
    setColumn( b, colA );
    setColumn( a, colB );
    
    if ( aa ) set( b, b );
    else unset( b, b );
    
    if ( ab ) set( b, a );
    else unset( b, a );
    
    if ( ba ) set( a, b );
    else unset( a, b );
    
    if ( bb ) set( a, a );
    else unset( a, a );
}

/*
 *  The following operators simply work on the operators of the
 *  underlying BitVectors.
 */

BitMatrix BitMatrix::operator ~() {
    BitMatrix res = *this;
    
    for ( unsigned int i = 0; i < height; ++i )
        res.rows[ i ] = ~(rows[ i ]);
    
    return res;
}

const BitMatrix& BitMatrix::operator &= ( const BitMatrix& other ) {
    assert( height == other.height );
    
    for ( unsigned int i = 0; i < height; ++i )
        rows[ i ] &= other.rows[ i ];
    
    return *this;
}


const BitMatrix& BitMatrix::operator |= ( const BitMatrix& other ) {
    assert( height == other.height );
    
    for ( unsigned int i = 0; i < height; ++i )
        rows[ i ] |= other.rows[ i ];
    
    return *this;
}

const BitMatrix& BitMatrix::operator ^= ( const BitMatrix& other ) {
    assert( height == other.height );
    
    for ( unsigned int i = 0; i < height; ++i )
        rows[ i ] ^= other.rows[ i ];
    
    return *this;
}


BitMatrix BitMatrix::operator & ( const BitMatrix& other ) {
    BitMatrix res( *this );
    
    res &= other;
    
    return res;
}


BitMatrix BitMatrix::operator | ( const BitMatrix& other ) {
    BitMatrix res( *this );
    
    res |= other;
    
    return res;
}

BitMatrix BitMatrix::operator ^ ( const BitMatrix& other ) {
    BitMatrix res( *this );
    
    res ^= other;
    
    return res;
}



BitMatrix& BitMatrix::operator = ( const BitMatrix& other) {
    
    if ( *this != other ) {
        if ( rows ) {
            delete[] rows;
        }
        
        height = other.height;
        width = other.width;
        
        rows = new BitVector[ height ];
        for ( unsigned int i = 0; i < height; ++i )
            rows[ i ] = other.rows[ i ];
    }
    
    return *this;
}


bool BitMatrix::operator ==( const BitMatrix& other ) const {
    if ( height != other.height ) return false;
    if ( width != other.width ) return false;
    
    for ( unsigned int i = 0; i < height; ++i ) {
        if ( rows[ i ] != other.rows[ i ] ) return false;
    }
    
    return true;
}

bool BitMatrix::operator !=( const BitMatrix& other ) const {
    return !( *this == other );
}



/*
 *  Debugging methods
 */

void BitMatrix::print() const {
    for ( int i = 0; i < height; ++i ) {
        std::cout << i << ( i > 9 ? "| " : " | ");
        rows[ i ].print();
    }
    std::cout << std::endl;
}
