#include "BitVector.h"
#define INT_SIZE 32

#include <assert.h>

#ifndef NULL
#define NULL 0
#endif

using namespace shared;

// create a bitvector of given size. 
BitVector::BitVector( const BitVector& other ) {
   bitcount = other.bitcount;
   length = other.length;
    
   vector = new int[ length ];
   for ( int i = 0; i < length; ++i ) {
       vector[ i ] = other.vector[ i ];
   }
}

BitVector::BitVector( unsigned int size ) {
    bitcount = size;
    length = (bitcount / INT_SIZE ) + ( bitcount % INT_SIZE ? 1 : 0);
    vector = new int[ length ];    
        
    for ( int i = 0; i < length; ++i )
        vector[ i ] = 0;
}

BitVector::BitVector() {
    bitcount = length = 0;
    vector = NULL;
}

BitVector::~BitVector() {
    if ( vector ) {
		delete[] vector;
    }
}

/*
 *  Basic modifiers for settings or unsetting of single bits, clearing the whole 
 *  vector or checking for single bits.
 */

/*!
*  It sets the bit at the given position to true(to 1).
* @param bit the position to be set to true.
*/
inline BitVector& BitVector::set( unsigned int bit ) {
   assert( bit < bitcount );
   vector[ bit / INT_SIZE ] |= 1 << (bit % INT_SIZE);
   
   return *this;
}

inline BitVector& BitVector::unset( unsigned int bit ) {
   assert( bit < bitcount ); 
   vector[ bit / (INT_SIZE)  ] &= ~(1 << (bit % INT_SIZE));
   
   return *this;
}

inline bool BitVector::isSet( unsigned int bit ) const {
    assert( bit < bitcount);
    return vector[ bit / (INT_SIZE)  ] & (1 << (bit % INT_SIZE) );
}

BitVector& BitVector::clear( bool value ) {
    int clearval = value ? ~0 : 0;

    for ( int i = 0; i < length; ++i )
        vector[ i ] = clearval;    
    
    // mask off all remaining bits of the last integer, so 
    // that it can be asserted that the unused bits are always set to
    // zero (else the comparison via the integers could fail! )
    if ( bitcount % INT_SIZE != 0 )
        vector[ length-1 ] &= (1 << (bitcount-((length-1)*INT_SIZE) ) )-1;
    
    return *this;
}

int BitVector::getChunk( unsigned int index ) const {
    assert( index < length );
    return vector[ index ];
}

BitVector& BitVector::setChunk( unsigned int index, int chunk ) {
    assert( index < length );
    
    vector[ index ] = chunk;
    
    // mask off unused bits
    if ( index == (length-1) && bitcount % INT_SIZE )
        vector[ length-1 ] &= (1 << (bitcount-((length-1)*INT_SIZE) ) )-1;
    
    return *this;
}


unsigned int BitVector::size() const {
    return bitcount;
}

unsigned int BitVector::sum() const {
    unsigned int res = 0;
    
    for( int i = 0; i < length; ++i ) {
	int s = vector[ i ];
        for ( int f = 0; f < INT_SIZE && s ; ++f ) {
            s &= s - 1;
	    res++;
        }
    }

    return res;
}


/*
 *  Advanced operations ("substring" etc. )
 */

BitVector BitVector::cut( unsigned int from, unsigned int width ) const {
    assert( from < bitcount );
    assert( from+width <= bitcount );
	 
    /*
	width is inclusive: the bit at 'from' is counted, too.
    */   
 
    BitVector res( width );
    
    // TODO: make efficient.
    for ( int i = 0; i < width; ++i ) {
        if ( this->isSet( i+from ) ) res.set( i );
    }
        
    return res;    
}

BitVector& BitVector::paste( unsigned int at, const BitVector& other ) {
    assert( at+other.bitcount-1 < bitcount  );

    for ( int i = 0; i < other.bitcount; ++i ) {
        if ( other.isSet( i ) ) 
            this->set( at+i );
        else
            this->unset( at+i );
    }
        
    return *this;    
}

BitVector BitVector::copy() const {
    BitVector res( *this );
    return res;
}

/*
 *  The following operators work just like you would expect them to do.
 */

BitVector BitVector::operator ~() {
    BitVector res = *this;
    
    for ( int i = 0; i < length; ++i )
       res.vector[ i ] = ~res.vector[ i ];
    
    // mask off all remaining bits of the last integer, so 
    // that it can be asserted that the unused bits are always set to
    // zero (else the comparison via the integers could fail! )
    if ( bitcount % INT_SIZE != 0 )
        res.vector[ length-1 ] &= (1 << (bitcount-((length-1)*INT_SIZE) ) )-1;
    
    return res;
}

BitVector& BitVector::operator = ( const BitVector& other ) {
    if ( &other != this) {
        length = other.length;
        bitcount = other.bitcount;

        if ( vector != 0 ) delete[] vector;

        vector = new int[ length ];
        for ( int i = 0; i < length; ++i )
            vector[ i ] = other.vector[ i ];
    }
    
    return *this;
}

const BitVector& BitVector::operator &= ( const BitVector& other ) {
    int arrayLen = length < other.length ? length : other.length;
    
    for ( int i = 0; i < arrayLen; ++i )
       vector[ i ] &= other.vector[ i ];
    
    return *this;
}

const BitVector& BitVector::operator |= ( const BitVector& other ) {
    int arrayLen = length < other.length ? length : other.length;
    
    for ( int i = 0; i < arrayLen; ++i )
       vector[ i ] |= other.vector[ i ];
    
    return *this;
}

const BitVector& BitVector::operator ^= ( const BitVector& other ) {
    int arrayLen = length < other.length ? length : other.length;
    
    for ( int i = 0; i < arrayLen; ++i )
       vector[ i ] ^= other.vector[ i ];
    
    return *this;
}

BitVector BitVector::operator & ( const BitVector& other ) {
    BitVector res = *this;

    res &= other;
        
    return res;
}
 
BitVector BitVector::operator | ( const BitVector& other ) {
    BitVector res = *this;

    res |= other;
        
    return res;
}

BitVector BitVector::operator ^ ( const BitVector& other ) {
    BitVector res = *this;

    res ^= other;
        
    return res;
}

bool BitVector::operator ==( const BitVector& other ) const {
    if ( bitcount != other.bitcount ) return false;
    
    for ( int i = 0; i < length; ++i ) {
        if ( vector[ i ] != other.vector[ i ] ) return false;
    }
    
    return true;
}


bool BitVector::operator !=( const BitVector& other ) const {
    return !( *this == other);
}





/*
 * Debugging methods
 */ 

// print the bitvector to the standard output (read from left to right)
void BitVector::print() const {
    bool flag = false;
    for ( int i = 0; i < length; ++i ) {
        unsigned int b = 1;
        for ( int j = 0; j < INT_SIZE; ++j ) {
            if ( !flag && (i*INT_SIZE)+j >= bitcount) {
                std::cout << "\x1b[31m";
                flag = true;
            }            
            
            std::cout << ((( vector[i] & b ) >> j ) ? "1" : ".") << " ";
            b = b << 1;
        } 
        std::cout << " ";
    }    
    std::cout << "\x1b[0m\n";
}
